from __future__ import annotations

import abc
import contextlib
import csv
import hashlib
import io
import os
import re
import shutil
import stat
import subprocess
import sys
import tempfile
import tokenize
import zipfile
from base64 import urlsafe_b64encode
from pathlib import Path
from typing import Any, BinaryIO, Generator, Mapping, NamedTuple, TextIO

from pdm.pep517 import __version__
from pdm.pep517._vendor.packaging import tags
from pdm.pep517._vendor.packaging.specifiers import SpecifierSet
from pdm.pep517.base import Builder
from pdm.pep517.exceptions import BuildError, PDMWarning
from pdm.pep517.utils import get_abi_tag, get_platform, show_warning

WHEEL_FILE_FORMAT = (
    """\
Wheel-Version: 1.0
Generator: pdm-pep517 %s
Root-Is-Purelib: {is_purelib}
Tag: {tag}
"""
    % __version__
)

PY_LIMITED_API_PATTERN = r"cp3\d"


class RecordEntry(NamedTuple):
    relpath: str
    hash_digest: str
    size: str


class WheelEntry(metaclass=abc.ABCMeta):
    # Fix the date time for reproducible builds
    date_time = (2016, 1, 1, 0, 0, 0)

    def __init__(self, rel_path: str) -> None:
        self.rel_path = rel_path

    @abc.abstractmethod
    def open(self) -> BinaryIO:
        pass

    def build_zipinfo(self) -> zipfile.ZipInfo:
        return zipfile.ZipInfo(self.rel_path, self.date_time)

    def write_to_zip(self, zf: zipfile.ZipFile) -> RecordEntry:
        zi = self.build_zipinfo()

        hashsum = hashlib.sha256()
        with self.open() as src:
            while True:
                buf = src.read(1024 * 8)
                if not buf:
                    break
                hashsum.update(buf)

            src.seek(0)
            zf.writestr(zi, src.read(), compress_type=zipfile.ZIP_DEFLATED)

        size = zi.file_size
        hash_digest = urlsafe_b64encode(hashsum.digest()).decode("ascii").rstrip("=")
        return RecordEntry(self.rel_path, f"sha256={hash_digest}", str(size))


class WheelFileEntry(WheelEntry):
    def __init__(self, rel_path: str, full_path: Path) -> None:
        super().__init__(rel_path)
        self.full_path = full_path

    def open(self) -> BinaryIO:
        return self.full_path.open("rb")

    def build_zipinfo(self) -> zipfile.ZipInfo:
        zi = super().build_zipinfo()
        st_mode = os.stat(self.full_path).st_mode
        zi.external_attr = (st_mode & 0xFFFF) << 16  # Unix attributes

        if stat.S_ISDIR(st_mode):
            zi.external_attr |= 0x10  # MS-DOS directory flag
        return zi


class WheelStringEntry(WheelEntry):
    def __init__(self, rel_path: str) -> None:
        super().__init__(rel_path)
        self.buffer = io.BytesIO()

    def open(self) -> BinaryIO:
        self.buffer.seek(0)
        return self.buffer

    @contextlib.contextmanager
    def text_open(self) -> Generator[TextIO, None, None]:
        text_buffer = io.TextIOWrapper(self.open(), encoding="utf-8", newline="")
        yield text_buffer
        text_buffer.detach()


class WheelBuilder(Builder):
    def __init__(
        self,
        location: str | Path,
        config_settings: Mapping[str, Any] | None = None,
    ) -> None:
        super().__init__(location, config_settings)
        self._entries: dict[str, WheelEntry] = {}
        self._parse_config_settings()

    def _parse_config_settings(self) -> None:
        self.python_tag = None
        self.py_limited_api = None
        self.plat_name = None
        if not self.config_settings:
            return
        if "--python-tag" in self.config_settings:
            self.python_tag = self.config_settings["--python-tag"]
        if "--py-limited-api" in self.config_settings:
            self.py_limited_api = self.config_settings["--py-limited-api"]
            if not re.match(PY_LIMITED_API_PATTERN, self.py_limited_api):
                raise ValueError(
                    "py-limited-api must match '%s'" % PY_LIMITED_API_PATTERN
                )
        if "--plat-name" in self.config_settings:
            self.plat_name = self.config_settings["--plat-name"]

    def build(
        self, build_dir: str, *, metadata_directory: str | None = None, **kwargs: Any
    ) -> str:
        if not os.path.exists(build_dir):
            os.makedirs(build_dir, exist_ok=True)

        self._entries.clear()
        fd, temp_path = tempfile.mkstemp(suffix=".whl")
        os.close(fd)

        self._copy_module()
        self._build()
        if metadata_directory is None:
            self._write_metadata()
        else:
            self._copy_metadata(metadata_directory)
        with zipfile.ZipFile(
            temp_path, mode="w", compression=zipfile.ZIP_DEFLATED
        ) as zip_file:
            self._write_to_zip(zip_file)

        target = os.path.join(build_dir, self.wheel_filename)
        if os.path.exists(target):
            os.unlink(target)
        shutil.move(temp_path, target)

        return target

    @property
    def wheel_filename(self) -> str:
        name = self.meta.project_filename
        version = self.meta_version
        return f"{name}-{version}-{self.tag}.whl"

    @property
    def tag(self) -> str:
        platform = self.plat_name
        impl = self.python_tag
        if not self.meta.config.is_purelib:
            if not platform:
                platform = get_platform(self.location / "build")
            if not impl:
                impl = tags.interpreter_name() + tags.interpreter_version()
            if self.py_limited_api and impl.startswith("cp3"):
                impl = self.py_limited_api
                abi_tag = "abi3"
            else:
                abi_tag = str(get_abi_tag()).lower()
        else:
            if not platform:
                platform = "any"
            abi_tag = "none"
            if not impl:
                if self.meta.requires_python and SpecifierSet(
                    self.meta.requires_python
                ).contains("2.7"):
                    impl = "py2.py3"
                else:
                    impl = "py3"

        platform = platform.lower().replace("-", "_").replace(".", "_")
        tag = (impl, abi_tag, platform)
        if not self.meta.config.is_purelib:
            supported_tags = [(t.interpreter, t.abi, platform) for t in tags.sys_tags()]
            assert (
                tag in supported_tags
            ), f"would build wheel with unsupported tag {tag}"
        return "-".join(tag)

    @property
    def dist_info_name(self) -> str:
        name = self.meta.project_filename
        version = self.meta_version
        return f"{name}-{version}.dist-info"

    def _write_record(self, records: list[RecordEntry]) -> WheelEntry:
        entry = WheelStringEntry(self.dist_info_name + "/RECORD")
        with entry.text_open() as fp:
            writer = csv.writer(fp, lineterminator="\n")
            writer.writerows(records)
            writer.writerow(RecordEntry(entry.rel_path, "", ""))
        return entry

    def _write_metadata(self) -> None:
        dist_info = self.dist_info_name
        if self.meta.entry_points:
            with self._open_for_write(dist_info + "/entry_points.txt") as f:
                self._write_entry_points(f)

        with self._open_for_write(dist_info + "/WHEEL") as f:
            self._write_wheel_file(f)

        with self._open_for_write(dist_info + "/METADATA") as f:
            self._write_metadata_file(f)

        for license_file in self.find_license_files():
            self._add_file(
                f"{dist_info}/licenses/{license_file}",
                self.location / license_file,
            )

    @contextlib.contextmanager
    def _open_for_write(self, rel_path: str) -> Generator[TextIO, None, None]:
        entry = WheelStringEntry(rel_path)
        with entry.text_open() as fp:
            yield fp
        print(f" - Adding {rel_path}")
        self._entries[rel_path] = entry

    def _build(self) -> None:
        build_dir = self.location / "build"
        if build_dir.exists():
            shutil.rmtree(str(build_dir))
        lib_dir: Path | None = None
        if self.meta.config.setup_script:
            if self.meta.config.run_setuptools:
                setup_py = self.ensure_setup_py()
                build_args = [
                    sys.executable,
                    str(setup_py),
                    "build",
                    "-b",
                    str(build_dir),
                ]
                try:
                    subprocess.check_call(build_args)
                except subprocess.CalledProcessError as e:
                    raise BuildError(f"Error occurs when running {build_args}:\n{e}")
                lib_dir = next(build_dir.glob("lib.*"), None)
            else:
                build_dir.mkdir(exist_ok=True)
                with tokenize.open(self.meta.config.setup_script) as f:
                    code = compile(f.read(), self.meta.config.setup_script, "exec")
                global_dict: dict[str, Any] = {}
                exec(code, global_dict)
                if "build" not in global_dict:
                    show_warning(
                        "No build() function found in the setup script, do nothing",
                        PDMWarning,
                    )
                    return
                global_dict["build"](str(self.location), str(build_dir))
                lib_dir = build_dir
        if lib_dir is None:
            lib_dir = build_dir
        if not lib_dir.exists():
            lib_dir.mkdir(parents=True)
        self._write_version(lib_dir)

        _, excludes = self._get_include_and_exclude_paths(for_sdist=False)
        for pkg in lib_dir.glob("**/*"):
            if pkg.is_dir():
                continue

            whl_path = rel_path = pkg.relative_to(lib_dir).as_posix()
            if self.meta.config.package_dir:
                # act like being in the package_dir
                rel_path = os.path.join(self.meta.config.package_dir, rel_path)

            if self._is_excluded(rel_path, excludes):
                continue

            self._add_file(whl_path, pkg)

    def _write_version(self, destination: Path) -> None:
        dynamic_version = self.meta.config.dynamic_version
        if (
            not dynamic_version
            or dynamic_version.source == "file"
            or "write_to" not in dynamic_version.options
        ):
            return
        write_template = dynamic_version.options.get("write_template", "{}\n")
        write_to = dynamic_version.options["write_to"]
        write_path = destination / write_to
        write_path.parent.mkdir(parents=True, exist_ok=True)
        with write_path.open("w") as f:
            f.write(write_template.format(self.meta_version))

    def _copy_module(self) -> None:
        root = self.meta.config.package_dir or self.location
        for path in self.find_files_to_add():
            try:
                rel_path = path.relative_to(root).as_posix()
            except ValueError:
                rel_path = path.as_posix()
            self._add_file(rel_path, path)

    def _add_file(self, rel_path: str, full_path: Path) -> None:
        if os.sep != "/":
            # We always want to have /-separated paths in the zip file and in RECORD
            rel_path = rel_path.replace(os.sep, "/")
        print(f" - Adding {rel_path}")
        self._entries[rel_path] = WheelFileEntry(rel_path, full_path)

    def _write_metadata_file(self, fp: TextIO) -> None:
        fp.write(self.format_pkginfo())

    def _write_wheel_file(self, fp: TextIO) -> None:
        fp.write(
            WHEEL_FILE_FORMAT.format(
                is_purelib=self.meta.config.is_purelib, tag=self.tag
            )
        )

    def _write_entry_points(self, fp: TextIO) -> None:
        entry_points = self.meta.entry_points
        for group_name in sorted(entry_points):
            fp.write(f"[{group_name}]\n")
            for ep in sorted(entry_points[group_name]):
                fp.write(ep.replace(" ", "") + "\n")

            fp.write("\n")

    def _write_to_zip(self, zf: zipfile.ZipFile) -> None:
        records: list[RecordEntry] = []
        for entry in self._entries.values():
            records.append(entry.write_to_zip(zf))

        record_entry = self._write_record(records)
        record_entry.write_to_zip(zf)

    def _copy_metadata(self, metadata_directory: str) -> None:
        for root, _, files in os.walk(metadata_directory):
            relroot = os.path.relpath(root, os.path.dirname(metadata_directory))
            for file in files:
                self._add_file(
                    os.path.join(relroot, file),
                    Path(root) / file,
                )



import io
import itertools
import os
import tarfile
import tempfile
from copy import copy
from typing import Any, Iterator

from pdm.pep517._vendor import tomli, tomli_w
from pdm.pep517.base import Builder


def normalize_file_permissions(st_mode: int) -> int:
    """
    Normalizes the permission bits in the st_mode field from stat to 644/755

    Popular VCSs only track whether a file is executable or not. The exact
    permissions can vary on systems with different umasks. Normalising
    to 644 (non executable) or 755 (executable) makes builds more reproducible.
    """
    # Set 644 permissions, leaving higher bits of st_mode unchanged
    new_mode = (st_mode | 0o644) & ~0o133
    if st_mode & 0o100:
        new_mode |= 0o111  # Executable: 644 -> 755

    return new_mode


def clean_tarinfo(tar_info: tarfile.TarInfo) -> tarfile.TarInfo:
    """
    Clean metadata from a TarInfo object to make it more reproducible.

        - Set uid & gid to 0
        - Set uname and gname to ""
        - Normalise permissions to 644 or 755
        - Set mtime if not None
    """
    ti = copy(tar_info)
    ti.uid = 0
    ti.gid = 0
    ti.uname = ""
    ti.gname = ""
    ti.mode = normalize_file_permissions(ti.mode)

    return ti


class SdistBuilder(Builder):
    """This build should be performed for PDM project only."""

    def _find_files_iter(self, for_sdist: bool = False) -> Iterator[str]:
        return itertools.chain(
            super()._find_files_iter(for_sdist), self.find_license_files()
        )

    def build(self, build_dir: str, **kwargs: Any) -> str:
        if not os.path.exists(build_dir):
            os.makedirs(build_dir, exist_ok=True)

        version = self.meta_version

        target = os.path.join(build_dir, f"{self.meta.project_name}-{version}.tar.gz")
        tar = tarfile.open(target, mode="w:gz", format=tarfile.PAX_FORMAT)

        try:
            tar_dir = f"{self.meta.project_name}-{version}"

            files_to_add = self.find_files_to_add(True)

            for relpath in files_to_add:
                if str(relpath) == "pyproject.toml":
                    self._add_pyproject(tar, tar_dir)
                else:
                    tar.add(
                        relpath,
                        arcname=os.path.join(tar_dir, str(relpath)),
                        recursive=False,
                    )
                print(f" - Adding {relpath}")

            fd, temp_name = tempfile.mkstemp(prefix="pkg-info")
            pkg_info = self.format_pkginfo(False).encode("utf-8")
            with open(fd, "wb") as f:
                f.write(pkg_info)
            tar.add(
                temp_name, arcname=os.path.join(tar_dir, "PKG-INFO"), recursive=False
            )
            print(" - Adding PKG-INFO")
        finally:
            tar.close()

        return target

    def _add_pyproject(self, tar: tarfile.TarFile, tar_dir: str) -> None:
        """Rewrites the pyproject.toml before adding to tarball.
        This is mainly aiming at fixing the version number in pyproject.toml
        """
        with open(self.location / "pyproject.toml", "rb") as f:
            pyproject = tomli.load(f)
        if self.meta.dynamic and "version" in self.meta.dynamic:
            self.meta.data["version"] = self.meta.version
            self.meta.data["dynamic"].remove("version")
        pyproject["project"] = self.meta.data
        name = "pyproject.toml"
        tarinfo = tar.gettarinfo(name, os.path.join(tar_dir, name))
        bio = io.BytesIO()
        tomli_w.dump(pyproject, bio)
        tarinfo.size = len(bio.getvalue())
        bio.seek(0)
        tar.addfile(tarinfo, bio)



from __future__ import annotations

import glob
import os
from pathlib import Path
from typing import Any, Callable, Generic, Iterable, Mapping, TypeVar, cast

from pdm.pep517._vendor.packaging.requirements import InvalidRequirement, Requirement
from pdm.pep517.exceptions import MetadataError, PDMWarning, ProjectError
from pdm.pep517.utils import (
    cd,
    expand_vars,
    find_packages_iter,
    merge_marker,
    safe_name,
    show_warning,
    to_filename,
)
from pdm.pep517.validator import validate_pep621
from pdm.pep517.version import DynamicVersion

T = TypeVar("T")


class MetaField(Generic[T]):
    def __init__(
        self, name: str, fget: Callable[[Metadata, Any], T] | None = None
    ) -> None:
        self.name = name
        self.fget = fget

    def __get__(self, instance: Metadata, owner: type[Metadata]) -> T | None:
        if instance is None:
            return self
        try:
            rv = instance.data[self.name]
        except KeyError:
            return None
        if self.fget is not None:
            rv = self.fget(instance, rv)
        return rv


class Metadata:
    """A class that holds all metadata that Python packaging requries."""

    DEFAULT_ENCODING = "utf-8"
    SUPPORTED_CONTENT_TYPES = ("text/markdown", "text/x-rst", "text/plain")

    def __init__(self, root: str | Path, pyproject: dict[str, Any]) -> None:
        self.root = Path(root).absolute()
        if "project" not in pyproject:
            raise ProjectError("No [project] config in pyproject.toml")
        self.data = pyproject["project"]
        self.config = Config(self.root, pyproject.get("tool", {}).get("pdm", {}))

    def validate(self, raising: bool = False) -> bool:
        return validate_pep621(self.data, raising)

    @property
    def name(self) -> str:
        if "name" not in self.data:
            raise MetadataError("name", "must be given in the project table")
        return self.data["name"]

    @property
    def version(self) -> str | None:
        static_version = self.data.get("version")
        if isinstance(static_version, str):
            return static_version
        if "version" in self.config.data and not (
            self.dynamic and "version" in self.dynamic
        ):
            raise MetadataError("version", "missing from 'dynamic' fields")
        dynamic_version = self.config.dynamic_version
        return (
            dynamic_version.evaluate_in_project(self.root) if dynamic_version else None
        )

    description: MetaField[str] = MetaField("description")

    def _get_readme_file(self, value: Mapping[str, str] | str) -> str:
        if isinstance(value, str):
            return value
        return value.get("file", "")

    def _get_readme_content(self, value: Mapping[str, str] | str) -> str:
        if isinstance(value, str):
            return Path(value).read_text(encoding=self.DEFAULT_ENCODING)
        if "file" in value and "text" in value:
            raise MetadataError(
                "readme",
                "readme table shouldn't specify both 'file' "
                "and 'text' at the same time",
            )
        if "text" in value:
            return value["text"]
        file_path = value.get("file", "")
        encoding = value.get("charset", self.DEFAULT_ENCODING)
        return Path(file_path).read_text(encoding=encoding)

    def _get_content_type(self, value: Mapping[str, str] | str) -> str:
        if isinstance(value, str):
            if value.lower().endswith(".md"):
                return "text/markdown"
            elif value.lower().endswith(".rst"):
                return "text/x-rst"
            return "text/plain"
        content_type = value.get("content-type")
        if not content_type:
            raise MetadataError(
                "readme", "'content-type' is missing in the readme table"
            )
        if content_type not in self.SUPPORTED_CONTENT_TYPES:
            raise MetadataError(
                "readme", f"Unsupported readme content-type: {content_type}"
            )
        return content_type

    readme: MetaField[str] = MetaField("readme", _get_readme_file)
    long_description: MetaField[str] = MetaField("readme", _get_readme_content)
    long_description_content_type: MetaField[str] = MetaField(
        "readme", _get_content_type
    )

    def _get_name(self, value: Iterable[Mapping[str, str]]) -> str:
        result = []
        for item in value:
            if "email" not in item and "name" in item:
                result.append(item["name"])
        return ",".join(result)

    def _get_email(self, value: Iterable[Mapping[str, str]]) -> str:
        result = []
        for item in value:
            if "email" not in item:
                continue
            email = (
                item["email"]
                if "name" not in item
                else "{name} <{email}>".format(**item)
            )
            result.append(email)
        return ",".join(result)

    author: MetaField[str] = MetaField("authors", _get_name)
    author_email: MetaField[str] = MetaField("authors", _get_email)
    maintainer: MetaField[str] = MetaField("maintainers", _get_name)
    maintainer_email: MetaField[str] = MetaField("maintainers", _get_email)

    @property
    def classifiers(self) -> list[str]:
        classifers = set(self.data.get("classifiers", []))

        if self.dynamic and "classifiers" in self.dynamic:
            show_warning(
                "`classifiers` no longer supports dynamic filling, "
                "please remove it from `dynamic` fields and manually "
                "supply all the classifiers",
                PDMWarning,
                stacklevel=2,
            )

        return sorted(classifers)

    keywords: MetaField[str] = MetaField("keywords")
    project_urls: MetaField[dict[str, str]] = MetaField("urls")

    @property
    def license_expression(self) -> str | None:
        if "license" in self.data:
            if isinstance(self.data["license"], str):
                return self.data["license"]
            # show_warning(
            #     "'license.text' subtable is deprecated in favor of string 'license' "
            #     "value",
            #     PDMWarning,
            #     stacklevel=2,
            # )
            if "text" in self.data["license"]:
                return self.data["license"]["text"]
        show_warning("'license' is missing", PDMWarning, stacklevel=2)
        return None

    @property
    def license_files(self) -> dict[str, list[str]]:
        subtable_files = None
        if (
            "license" in self.data
            and isinstance(self.data["license"], dict)
            and "files" in self.data["license"]
        ):
            subtable_files = self.data["license"]["files"]
        if "license-files" not in self.data:
            if subtable_files is not None:
                # show_warning(
                #     "'license.files' subtable is deprecated in favor of "
                #     "'license-files'",
                #     PDMWarning,
                #     stacklevel=2,
                # )
                return {"paths": [self.data["license"]["file"]]}
            return {
                "globs": [
                    "LICENSES/*",
                    "LICEN[CS]E*",
                    "COPYING*",
                    "NOTICE*",
                    "AUTHORS*",
                ]
            }
        if subtable_files is not None:
            raise MetadataError(
                "license-files",
                "Can't specify both 'license.files' and 'license-files' fields",
            )
        rv = self.data["license-files"]
        valid_keys = {"globs", "paths"} & set(rv)
        if len(valid_keys) == 2:
            raise MetadataError(
                "license-files", "Can't specify both 'paths' and 'globs'"
            )
        if not valid_keys:
            raise MetadataError("license-files", "Must specify 'paths' or 'globs'")
        return rv

    def _convert_dependencies(
        self, deps: list[str], field: str = "dependencies"
    ) -> list[str]:
        def convert_req(req: str) -> str | None:
            """Discard all non-PEP 440 requirements, e.g. editable VCS requirements."""

            if req.strip().startswith("-e"):
                return None
            try:
                r = Requirement(req)
            except InvalidRequirement as e:
                raise MetadataError(field, f"Invalid requirement {req!r}\n  {e}") from e
            if r.url and "${" in r.url:
                r.url = expand_vars(r.url, self.root.resolve().as_posix())
            return str(r)

        return list(filter(None, map(convert_req, deps)))

    def _convert_optional_dependencies(
        self, deps: Mapping[str, list[str]]
    ) -> dict[str, list[str]]:
        return {
            k: self._convert_dependencies(deps[k], "optional-dependencies")
            for k in deps
        }

    dependencies: MetaField[list[str]] = MetaField(
        "dependencies", _convert_dependencies
    )
    optional_dependencies: MetaField[dict[str, list[str]]] = MetaField(
        "optional-dependencies", _convert_optional_dependencies
    )
    dynamic: MetaField[list[str]] = MetaField("dynamic")

    @property
    def project_name(self) -> str | None:
        if self.name is None:
            return None
        return safe_name(self.name)

    @property
    def project_filename(self) -> str:
        if self.project_name is None:
            return "UNKNOWN"
        return to_filename(self.project_name)

    @property
    def requires_extra(self) -> dict[str, list[str]]:
        """For PKG-INFO metadata"""
        if not self.optional_dependencies:
            return {}
        result: dict[str, list[str]] = {}
        for name, reqs in self.optional_dependencies.items():
            current = result[name] = []
            for r in reqs:
                parsed = Requirement(r)
                merge_marker(parsed, f"extra == {name!r}")
                current.append(str(parsed))
        return result

    @property
    def requires_python(self) -> str:
        result = self.data.get("requires-python", "")
        return "" if result == "*" else result

    @property
    def entry_points(self) -> dict[str, list[str]]:
        result = {}
        settings = self.data
        if "scripts" in settings:
            result["console_scripts"] = [
                f"{key} = {value}" for key, value in settings["scripts"].items()
            ]
        if "gui-scripts" in settings:
            result["gui_scripts"] = [
                f"{key} = {value}" for key, value in settings["gui-scripts"].items()
            ]
        if "entry-points" in settings:
            for plugin, value in settings["entry-points"].items():
                if plugin in ("console_scripts", "gui_scripts"):
                    correct_key = (
                        "scripts" if plugin == "console_scripts" else "gui-scripts"
                    )
                    raise MetadataError(
                        "entry-points",
                        f"entry-points {plugin!r} should be defined in "
                        f"[project.{correct_key}]",
                    )
                result[plugin] = [f"{k} = {v}" for k, v in value.items()]
        return result

    def convert_package_paths(self) -> dict[str, list | dict]:
        """Return a {package_dir, packages, package_data, exclude_package_data} dict."""
        packages = []
        py_modules = []
        package_data = {"": ["*"]}
        exclude_package_data: dict[str, list[str]] = {}
        package_dir = self.config.package_dir
        includes = self.config.includes
        excludes = self.config.excludes

        with cd(self.root.as_posix()):
            src_dir = Path(package_dir or ".")
            if not includes:
                packages = list(
                    find_packages_iter(
                        package_dir or ".",
                        exclude=["tests", "tests.*"],
                        src=str(src_dir),
                    )
                )
                if not packages:
                    py_modules = [path.name[:-3] for path in src_dir.glob("*.py")]
            else:
                packages_set = set()
                includes = includes[:]
                for include in includes[:]:
                    if include.replace("\\", "/").endswith("/*"):
                        include = include[:-2]
                    if "*" not in include and os.path.isdir(include):
                        dir_name = include.rstrip("/\\")
                        temp = list(
                            find_packages_iter(dir_name, src=package_dir or ".")
                        )
                        if os.path.isfile(os.path.join(dir_name, "__init__.py")):
                            temp.insert(0, dir_name)
                        packages_set.update(temp)
                        includes.remove(include)
                packages[:] = list(packages_set)
                for include in includes:
                    for path in glob.glob(include, recursive=True):
                        if "/" not in path.lstrip("./") and path.endswith(".py"):
                            # Only include top level py modules
                            py_modules.append(path.lstrip("./")[:-3])
                    if include.endswith(".py"):
                        continue
                    for package in packages:
                        relpath = os.path.relpath(include, package)
                        if not relpath.startswith(".."):
                            package_data.setdefault(package, []).append(relpath)
                for exclude in excludes:
                    for package in packages:
                        relpath = os.path.relpath(exclude, package)
                        if not relpath.startswith(".."):
                            exclude_package_data.setdefault(package, []).append(relpath)
            if packages and py_modules:
                raise ProjectError(
                    "Can't specify packages and py_modules at the same time."
                )
        return {
            "package_dir": {"": package_dir} if package_dir else {},
            "packages": packages,
            "py_modules": py_modules,
            "package_data": package_data,
            "exclude_package_data": exclude_package_data,
        }


class Config:
    """The [tool.pdm] table"""

    def __init__(self, root: Path, data: dict[str, Any]) -> None:
        self.root = root
        self.data = data

    def _compatible_get(
        self, name: str, default: Any = None, old_name: str | None = None
    ) -> Any:
        if name in self.data.get("build", {}):
            return self.data["build"][name]
        old_name = old_name or name
        if old_name in self.data:
            show_warning(
                f"Field `{old_name}` is renamed to `{name}` under [tool.pdm.build] "
                "table, please update your pyproject.toml accordingly",
                PDMWarning,
                stacklevel=2,
            )
            return self.data[old_name]
        return default

    @property
    def includes(self) -> list[str]:
        return self._compatible_get("includes", [])

    @property
    def source_includes(self) -> list[str]:
        return self._compatible_get("source-includes", [])

    @property
    def excludes(self) -> list[str]:
        return self._compatible_get("excludes", [])

    @property
    def setup_script(self) -> str | None:
        build_table = self.data.get("build", {})
        if "setup-script" in build_table:
            return build_table["setup-script"]
        if isinstance(build_table, str):
            show_warning(
                "Field `build` is renamed to `setup-script` under [tool.pdm.build] "
                "table, please update your pyproject.toml accordingly",
                PDMWarning,
                stacklevel=2,
            )
            return build_table
        return None

    @property
    def run_setuptools(self) -> bool:
        build = self.data.get("build", {})
        return isinstance(build, str) or cast(dict, build).get("run-setuptools", False)

    @property
    def package_dir(self) -> str:
        """A directory that will be used to looking for packages."""
        default = "src" if self.root.joinpath("src").exists() else ""
        return self._compatible_get("package-dir", default)

    @property
    def is_purelib(self) -> bool:
        """If not explicitly set, the project is considered to be non-pure
        if `build` exists.
        """
        return self._compatible_get("is-purelib", not bool(self.run_setuptools))

    @property
    def editable_backend(self) -> str:
        """Currently only two backends are supported:
        - editables: Proxy modules via editables
        - path: the legacy .pth file method(default)
        """
        return self._compatible_get("editable-backend", "path")

    @property
    def dynamic_version(self) -> DynamicVersion | None:
        dynamic_version = self.data.get("version")
        if not dynamic_version:
            return None
        return DynamicVersion.from_toml(dynamic_version)



"""
PEP-517 compliant buildsystem API
"""
from __future__ import annotations

import shutil
from pathlib import Path
from typing import Any, Mapping

from pdm.pep517.editable import EditableBuilder
from pdm.pep517.sdist import SdistBuilder
from pdm.pep517.wheel import WheelBuilder


def get_requires_for_build_wheel(
    config_settings: Mapping[str, Any] | None = None
) -> list[str]:
    """
    Returns an additional list of requirements for building, as PEP508 strings,
    above and beyond those specified in the pyproject.toml file.

    When C-extension build is needed, setuptools should be required, otherwise
    just return an empty list.
    """
    with WheelBuilder(Path.cwd(), config_settings) as builder:
        if builder.meta.config.run_setuptools:
            return ["setuptools>=40.8.0"]
        return []


def get_requires_for_build_sdist(
    config_settings: Mapping[str, Any] | None = None
) -> list[str]:
    """There isn't any requirement for building a sdist at this point."""
    return []


def _prepare_metadata(builder: WheelBuilder, metadata_directory: str) -> str:
    dist_info = Path(metadata_directory, builder.dist_info_name)
    dist_info.mkdir(exist_ok=True)
    if builder.meta.entry_points:
        with (dist_info / "entry_points.txt").open("w", encoding="utf-8") as f:
            builder._write_entry_points(f)

    with (dist_info / "WHEEL").open("w", encoding="utf-8") as f:
        builder._write_wheel_file(f)

    with (dist_info / "METADATA").open("w", encoding="utf-8") as f:
        builder._write_metadata_file(f)

    for license_file in builder.find_license_files():
        full_path = dist_info / "licenses" / license_file
        full_path.parent.mkdir(exist_ok=True, parents=True)
        shutil.copy2(license_file, full_path)

    return dist_info.name


def prepare_metadata_for_build_wheel(
    metadata_directory: str, config_settings: Mapping[str, Any] | None = None
) -> str:
    """Prepare the metadata, places it in metadata_directory"""
    with WheelBuilder(Path.cwd(), config_settings) as builder:
        return _prepare_metadata(builder, metadata_directory)


def build_wheel(
    wheel_directory: str,
    config_settings: Mapping[str, Any] | None = None,
    metadata_directory: str | None = None,
) -> str:
    """Builds a wheel, places it in wheel_directory"""
    with WheelBuilder(Path.cwd(), config_settings) as builder:
        return Path(
            builder.build(wheel_directory, metadata_directory=metadata_directory)
        ).name


def build_sdist(
    sdist_directory: str, config_settings: Mapping[str, Any] | None = None
) -> str:
    """Builds an sdist, places it in sdist_directory"""
    with SdistBuilder(Path.cwd(), config_settings) as builder:
        return Path(builder.build(sdist_directory)).name


get_requires_for_build_editable = get_requires_for_build_wheel


def prepare_metadata_for_build_editable(
    metadata_directory: str, config_settings: Mapping[str, Any] | None = None
) -> str:
    """Prepare the metadata, places it in metadata_directory"""
    with EditableBuilder(Path.cwd(), config_settings) as builder:
        builder._prepare_editable()
        return _prepare_metadata(builder, metadata_directory)


def build_editable(
    wheel_directory: str,
    config_settings: Mapping[str, Any] | None = None,
    metadata_directory: str | None = None,
) -> str:
    with EditableBuilder(Path.cwd(), config_settings) as builder:
        return Path(
            builder.build(wheel_directory, metadata_directory=metadata_directory)
        ).name
